import { __decorate } from "tslib";
import { CSSDirective, defaultExecutionContext, FASTElement, observable, Observable, } from "@microsoft/fast-element";
import { DI, Registration } from "../di/di";
import { CustomPropertyManager } from "./custom-property-manager";
/**
 * Implementation of {@link (DesignToken:interface)}
 */
class DesignTokenImpl extends CSSDirective {
    constructor(name) {
        super();
        this.name = name;
        this.customPropertyChangeHandlers = new Map();
        this.cssCustomProperty = `--${name}`;
        this.cssVar = `var(${this.cssCustomProperty})`;
    }
    createCSS() {
        return this.cssVar;
    }
    getValueFor(element) {
        const node = DesignTokenNode.for(this, element);
        Observable.track(node, "value");
        return DesignTokenNode.for(this, element).value;
    }
    setValueFor(element, value) {
        if (value instanceof DesignTokenImpl) {
            const _value = value;
            value = ((_element) => DesignTokenNode.for(_value, _element)
                .value);
        }
        DesignTokenNode.for(this, element).set(value);
        return this;
    }
    deleteValueFor(element) {
        DesignTokenNode.for(this, element).delete();
        return this;
    }
    addCustomPropertyFor(element) {
        if (!this.customPropertyChangeHandlers.has(element)) {
            const node = DesignTokenNode.for(this, element);
            let value = node.value;
            const add = () => CustomPropertyManager.addTo(element, this, value);
            const remove = () => CustomPropertyManager.removeFrom(element, this, value);
            const subscriber = {
                handleChange: (source, key) => {
                    remove();
                    value = source[key];
                    add();
                },
                dispose: () => {
                    remove();
                    Observable.getNotifier(node).unsubscribe(subscriber, "value");
                    this.customPropertyChangeHandlers.delete(element);
                },
            };
            this.customPropertyChangeHandlers.set(element, subscriber);
            add();
            Observable.getNotifier(node).subscribe(subscriber, "value");
        }
        return this;
    }
    removeCustomPropertyFor(element) {
        if (this.customPropertyChangeHandlers.has(element)) {
            this.customPropertyChangeHandlers.get(element).dispose();
        }
        return this;
    }
    createBehavior() {
        return new DesignTokenBehavior(this);
    }
}
/**
 * Behavior to add and Design Token custom properties for an element
 */
class DesignTokenBehavior {
    constructor(token) {
        this.token = token;
    }
    bind(target) {
        this.token.addCustomPropertyFor(target);
    }
    unbind(target) {
        this.token.removeCustomPropertyFor(target);
    }
}
const nodeCache = new WeakMap();
const channelCache = new Map();
const childToParent = new WeakMap();
const noop = Function.prototype;
/**
 * A node responsible for setting and getting token values,
 * emitting values to CSS custom properties, and maintaining
 * inheritance structures.
 */
class DesignTokenNode {
    constructor(token, target) {
        var _a;
        this.token = token;
        this.target = target;
        /** Track downstream nodes */
        this.children = new Set();
        /**
         * Invoked when parent node's value changes
         */
        this.handleChange = this.valueChangeHandler;
        if (nodeCache.has(target) && nodeCache.get(target).has(token)) {
            throw new Error(`DesignTokenNode already created for ${token.name} and ${target}. Use DesignTokenNode.for() to ensure proper reuse`);
        }
        const container = DI.getOrCreateDOMContainer(this.target);
        const channel = DesignTokenNode.channel(token);
        container.register(Registration.instance(channel, this));
        if (target instanceof FASTElement) {
            target.$fastController.addBehaviors([
                {
                    bind: () => { var _a; return (_a = this.findParentNode()) === null || _a === void 0 ? void 0 : _a.appendChild(this); },
                    unbind: () => { var _a; return (_a = childToParent.get(this)) === null || _a === void 0 ? void 0 : _a.removeChild(this); },
                },
            ]);
        }
        else {
            (_a = this.findParentNode()) === null || _a === void 0 ? void 0 : _a.appendChild(this);
        }
    }
    _rawValueChanged() {
        Observable.getNotifier(this).notify("value");
    }
    resolveRealValueForNode(node) {
        let current = node;
        while (current !== undefined) {
            if (current.rawValue !== undefined) {
                const { rawValue } = current;
                if (DesignTokenNode.isDerivedTokenValue(rawValue)) {
                    this.setupBindingObserver(rawValue);
                    return this.bindingObserver.observe(this.target, defaultExecutionContext);
                }
                else {
                    return rawValue;
                }
            }
            current = childToParent.get(current);
        }
        throw new Error(`Value could not be retrieved for token named "${this.token.name}". Ensure the value is set for ${this.target} or an ancestor of ${this.target}.`);
    }
    static channel(token) {
        return channelCache.has(token)
            ? channelCache.get(token)
            : channelCache.set(token, DI.createInterface()) &&
                channelCache.get(token);
    }
    static isDerivedTokenValue(value) {
        return typeof value === "function";
    }
    valueChangeHandler(source, key) {
        if (this._rawValue === void 0) {
            Observable.getNotifier(this).notify("value");
        }
    }
    setupBindingObserver(value) {
        this.tearDownBindingObserver();
        const handler = {
            handleChange: (source) => {
                Observable.getNotifier(this).notify("value");
            },
        };
        this.bindingObserver = Observable.binding(value, handler);
    }
    tearDownBindingObserver() {
        if (this.bindingObserver) {
            this.bindingObserver.disconnect();
            this.bindingObserver = undefined;
        }
    }
    static for(token, target) {
        const targetCache = nodeCache.has(target)
            ? nodeCache.get(target)
            : nodeCache.set(target, new Map()) && nodeCache.get(target);
        return targetCache.has(token)
            ? targetCache.get(token)
            : targetCache.set(token, new DesignTokenNode(token, target)) &&
                targetCache.get(token);
    }
    appendChild(child) {
        if (this.children.has(child)) {
            return;
        }
        this.children.forEach(c => {
            if (child.contains(c)) {
                this.removeChild(c);
                child.appendChild(c);
            }
        });
        this.children.add(child);
        Observable.getNotifier(this).subscribe(child, "value");
        childToParent.set(child, this);
    }
    removeChild(child) {
        this.children.delete(child);
        childToParent.delete(child);
        Observable.getNotifier(this).unsubscribe(child, "value");
    }
    contains(node) {
        return this.target.contains(node.target);
    }
    findParentNode() {
        if (this.target !== document.body && this.target.parentNode) {
            const container = DI.getOrCreateDOMContainer(this.target.parentElement);
            try {
                return container.get(DesignTokenNode.channel(this.token));
            }
            catch (e) {
                return null;
            }
        }
        else {
            return null;
        }
    }
    /**
     * The resolved value for a node.
     */
    get value() {
        try {
            return this.resolveRealValueForNode(this);
        }
        catch (e) {
            if (!childToParent.has(this)) {
                const parent = this.findParentNode();
                if (parent) {
                    parent.appendChild(this);
                    return this.resolveRealValueForNode(this);
                }
            }
            throw e;
        }
    }
    /**
     * The actual value set for the node, or undefined.
     * This will be a reference to the original object for all data types
     * passed by reference.
     */
    get rawValue() {
        return this._rawValue;
    }
    /**
     * Sets a value for the node
     * @param value The value to set
     */
    set(value) {
        if (value === this._rawValue) {
            return;
        }
        if (DesignTokenNode.isDerivedTokenValue(value)) {
            this.handleChange = noop;
            this.setupBindingObserver(value);
        }
        else if (this._rawValue !== value) {
            this.handleChange = noop;
        }
        this._rawValue = value;
    }
    /**
     * Deletes any value set for the node.
     */
    delete() {
        this._rawValue = void 0;
        this.handleChange = this.valueChangeHandler;
        this.tearDownBindingObserver();
    }
}
__decorate([
    observable
], DesignTokenNode.prototype, "_rawValue", void 0);
function create(name) {
    return new DesignTokenImpl(name);
}
/**
 * Factory object for creating {@link (DesignToken:interface)} instances.
 * @alpha
 */
export const DesignToken = Object.freeze({
    create,
});
